<?php

/**
 * ownCloud
 *
 * @author Artur Neumann <artur@jankaritech.com>
 * @copyright Copyright (c) 2017 Artur Neumann artur@jankaritech.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU AFFERO GENERAL PUBLIC LICENSE for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

namespace Page;

use Behat\Mink\Element\NodeElement;
use SensioLabs\Behat\PageObjectExtension\PageObject\Exception\ElementNotFoundException;
use WebDriver\Key;

/**
 * Page Object for a Firewall Rule
 * to keep the page objects small we do not define the whole page as one object
 * but logical parts of the page
 *
 */
class FileFirewallRule extends OwncloudPage {

	/**
	 * @var NodeElement of this rule
	 */
	protected $ruleElement;
	protected $filter;
	
	/**
	 * current rule type
	 *
	 * @var string (regex|normal)
	 */
	protected $type = 'normal';
	
	/**
	 * associated array of xpath to find the filter select drop-down
	 * one value for the "normal" case one value for the "regex" case
	 *
	 * @var array
	 */
	protected $filterSelectXpath = [
			'normal' => '//select[contains(@name,"filter")]',
			'regex' => './/select[@class="regexRules"]'
	];
	/**
	 * type element xpath. type element and filter element are currently the same.
	 * we still want to have separate variables for it as that might change
	 *
	 * @var string
	 */
	protected $typeSelectXpath = '//select[contains(@name,"filter")]';
	protected $operatorSelectXpath = '//select[contains(@name,"operator")]';
	
	/**
	 * associated array of xpath to find the value field (select or input)
	 * one value for the "normal" case one value for the "regex" case
	 *
	 * @var array
	 */
	protected $valueInputXpath = [
			'normal' => '//input[contains(@name,"value")]',
			'regex' => '//input[@name="regexValue"]'
	];
	protected $requestTimeFilterName = "Request Time";
	protected $requestTimeInputTimeFormat = "h:i a O";
	protected $tagsSelectContainerXpath
		= './/*[contains(@class,"select2-container")]';
	protected $tagFilterName = "System file tag";
	protected $tagInputFieldSuffix = "_search";
	protected $regularExpressionOptionTitle = "Regular Expression";
	protected $emptyFilterValue = "-1";
	
	/**
	 * array of dropdown select fields "displayed text" => "css class"
	 *
	 * @var array
	 */
	protected $dropdownValues = [
			"User Device" => "userDevices",
			"Request Type" => 'requestType',
			"User Group" => "userGroups"
	];
	protected $dropdownValueSelectXpath = ".//select[@class='%s']";
	protected $deleteBtnXpath = ".//button[@data-delete='rule']";
	protected $daterPickerCloseBtnXpath
		= '//button[contains(@class,"ui-datepicker-close")]';

	/**
	 * see {@link FileFirewallGroup::setElement($element)}
	 *
	 * @param NodeElement $ruleElement
	 *
	 * @return void
	 */
	public function setElement(NodeElement $ruleElement) {
		$this->ruleElement = $ruleElement;
	}

	/**
	 * @return void
	 * @throws ElementNotFoundException
	 */
	public function delete() {
		$deleteBtn = $this->ruleElement->find("xpath", $this->deleteBtnXpath);
		if ($deleteBtn === null) {
			throw new ElementNotFoundException("could not find delete button");
		}
		$deleteBtn->click();
	}

	/**
	 *
	 * @return boolean
	 */
	public function isEmpty() {
		if ($this->getFilterValue() === $this->emptyFilterValue) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 *
	 * @throws ElementNotFoundException
	 * @return NodeElement
	 */
	public function findFilterElement() {
		$filterSelect = $this->ruleElement->find(
			"xpath", $this->filterSelectXpath[$this->type]
		);
		if ($filterSelect === null) {
			throw new ElementNotFoundException(
				"could not find filter select field"
			);
		}
		return $filterSelect;
	}

	/**
	 *
	 * @throws ElementNotFoundException
	 * @return NodeElement
	 */
	public function findTypeElement() {
		$typeSelect = $this->ruleElement->find("xpath", $this->typeSelectXpath);
		if ($typeSelect === null) {
			throw new ElementNotFoundException("could not find type select field");
		}
		return $typeSelect;
	}

	/**
	 *
	 * @return string
	 */
	public function getFilterValue() {
		return $this->findFilterElement()->getValue();
	}

	/**
	 *
	 * @throws ElementNotFoundException
	 * @return NodeElement
	 */
	public function findOperatorElement() {
		$operatorSelect = $this->ruleElement->find(
			"xpath", $this->operatorSelectXpath
		);
		if ($operatorSelect === null) {
			throw new ElementNotFoundException(
				"could not find operator select field"
			);
		}
		return $operatorSelect;
	}

	/**
	 *
	 * @return string
	 */
	public function getOperatorValue() {
		return $this->findOperatorElement()->getValue();
	}

	/**
	 *
	 * @param string $type (regex|normal)
	 *
	 * @return void
	 */
	public function setType($type) {
		if ($type === "regex") {
			$this->findTypeElement()->selectOption(
				$this->regularExpressionOptionTitle
			);
		}
		$this->type = $type;
	}

	/**
	 *
	 * @param string $option human visible text, NOT value
	 *
	 * @return void
	 */
	public function selectFilter($option) {
		$this->findFilterElement()->selectOption($option);
		$this->filter = $option;
	}

	/**
	 *
	 * @param string $option human visible text, NOT value
	 *
	 * @return void
	 */
	public function selectOperator($option) {
		$this->findOperatorElement()->selectOption($option);
	}

	/**
	 *
	 * @throws ElementNotFoundException
	 * @return NodeElement[] all input fields
	 */
	public function findValueElement() {
		if ($this->filter === $this->tagFilterName) {
			//tags filter are special so need to find them in a special way
			$tagsSelectContainer = $this->find(
				"xpath", $this->tagsSelectContainerXpath
			);
			if ($tagsSelectContainer === null) {
				throw new ElementNotFoundException(
					"could not find tags select container"
				);
			}
			$tagsSelectContainer->click();
			$inputId = $tagsSelectContainer->find("xpath", "//input")
				->getAttribute("id") . $this->tagInputFieldSuffix;
			$valueInput[0] = $this->findById($inputId);
		} elseif (isset($this->dropdownValues[$this->filter])
			&& $this->type === "normal"
		) {
			//any of the "normal" dropdown select elements
			$valueInput[0] = $this->ruleElement->find(
				"xpath",
				\sprintf(
					$this->dropdownValueSelectXpath,
					$this->dropdownValues[$this->filter]
				)
			);
		} else {
			//all other input fields are easy
			$valueInput = $this->ruleElement->findAll(
				"xpath", $this->valueInputXpath[$this->type]
			);
		}

		if (!isset($valueInput[0]) || $valueInput[0] === null) {
			throw new ElementNotFoundException("could not find value input field");
		}
		return $valueInput;
	}

	/**
	 *
	 * @param string $value
	 *
	 * @return void
	 */
	public function setValue($value) {
		//the "System file tag" does work different,
		//so we need to send an ENTER key at the end
		if ($this->filter === $this->tagFilterName) {
			$value = $value . Key::ENTER;
		}
		if ($this->filter === $this->requestTimeFilterName) {
			//request time needs to set 2 values
			$time = \explode(",", $value);
			$this->findValueElement() [0]->setValue(
				\date($this->requestTimeInputTimeFormat, \strtotime($time [0]))
				. Key::ENTER
			);
			$this->findValueElement()[1]->setValue(
				\date($this->requestTimeInputTimeFormat, \strtotime($time [1]))
				. Key::ENTER
			);
			// close the datepicker if its still open,
			// if it doesn't work ignore the problem
			try {
				$this->find("xpath", $this->daterPickerCloseBtnXpath)->click();
			} catch (\Exception $e) {
				//ignore
			}
		} elseif (isset($this->dropdownValues[$this->filter])
			&& $this->type === "normal"
		) {
			//all "normal" dropdown select elements
			$this->findValueElement()[0]->selectOption($value);
		} else {
			$this->findValueElement()[0]->setValue($value);
		}
	}
}
