<?php
/**
 * ownCloud Firewall
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright Copyright (c) 2018, ownCloud GmbH
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Firewall\Tests\AppInfo;

use OC\Server;
use OCA\Firewall\AppInfo\Application;
use OCA\Firewall\Config;
use OCP\App\IAppManager;
use OCP\AppFramework\IAppContainer;
use OCP\ILogger;
use OCP\License\ILicenseManager;
use Test\TestCase;

/**
 * Class ApplicationTest
 *
 * @group DB
 * @package OCA\Firewall\Tests\AppInfo
 */
class ApplicationTest extends TestCase {
	/** @var Application */
	protected $application;

	protected function setUp(): void {
		parent::setUp();

		$this->application = new Application();
	}

	public function dataContainerQuery() {
		return [
			['OCA\Firewall\Config', '\OCA\Firewall\Config'],
			['OCA\Firewall\Context', '\OCA\Firewall\Context'],
			['OCA\Firewall\Logger', '\OCA\Firewall\Logger'],
			['OCA\Firewall\RuleFactory', '\OCA\Firewall\RuleFactory'],
			['OCA\Firewall\Ruler', '\OCA\Firewall\Ruler'],
			['OCA\Firewall\Firewall', '\OCA\Firewall\Firewall'],
		];
	}

	/**
	 * @dataProvider dataContainerQuery
	 *
	 * @param string $query
	 * @param string $class
	 */
	public function testContainerQuery($query, $class) {
		$object = $this->application->getContainer()->query($query);
		$this->assertInstanceOf($class, $object);
	}

	public function dataBoot() {
		return [
			[[], false, false],
			[[], true, false],
			[['not_empty'], false, false],
			[['not_empty'], true, true],
		];
	}

	/**
	 * @dataProvider dataBoot
	 *
	 * @param array $rules
	 * @param bool $setUpFS
	 * @param bool $wrapStorage
	 */
	public function testBoot(array $rules, $setUpFS, $wrapStorage) {
		/** @var \OCA\Firewall\AppInfo\Application|\PHPUnit\Framework\MockObject\MockObject $application */
		$application = $this->getMockBuilder('OCA\Firewall\AppInfo\Application')
			->disableOriginalConstructor()
			->setMethods([
				'getContainer',
				'wrapStorage',
			])
			->getMock();

		$config = $this->createMock(Config::class);
		$config->expects($this->once())
			->method('getRules')
			->with(true)
			->willReturn($rules);

		$licenseManager = $this->createMock(ILicenseManager::class);
		$licenseManager->method('checkLicenseFor')
			->with('firewall')
			->willReturn(true);
		$server = $this->createMock(Server::class);
		$server->method('getLicenseManager')
			->willReturn($licenseManager);
		$container = $this->createMock(IAppContainer::class);
		$container->method('query')
			->withConsecutive(['ServerContainer'], [Config::class])
			->willReturnOnConsecutiveCalls($server, $config);

		if ($wrapStorage) {
			$application->expects($this->once())
				->method('wrapStorage')
				->willReturnArgument(1);
		} else {
			$application->expects($this->never())
				->method('wrapStorage');
		}

		$application->expects($this->once())
			->method('getContainer')
			->willReturn($container);

		$application->boot();

		if ($setUpFS) {
			\OCP\Util::emitHook('OC_Filesystem', 'preSetup', []);
		}
	}

	public function dataWrapper() {
		return [
			[false, false, true],
			[true, false, false],
			[false, true, false],
			[true, true, false],
		];
	}

	/**
	 * @dataProvider dataWrapper
	 *
	 * @param bool $isShared
	 * @param bool $isCli
	 * @param bool $wrapStorage
	 */
	public function testWrapper($isShared, $isCli, $wrapStorage) {
		$application = new \OCA\Firewall\AppInfo\Application();

		/** @var \OCP\Files\Storage|\PHPUnit\Framework\MockObject\MockObject $storage */
		$storage = $this->getMockBuilder('OCP\Files\Storage')
			->disableOriginalConstructor()
			->getMock();

		$storage->expects($this->once())
			->method('instanceOfStorage')
			->with('OC\Files\Storage\Shared')
			->willReturn($isShared);

		$prevCLI = \OC::$CLI;
		\OC::$CLI = $isCli;

		$return = $application->wrapper('mountPoint', $storage);

		\OC::$CLI = $prevCLI;

		if ($wrapStorage) {
			$this->assertInstanceOf('OCA\Firewall\StorageWrapper', $return);
		} else {
			$this->assertSame($storage, $return);
		}
	}
}
