<?php
/**
 * ownCloud Workflow
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright 2015 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Workflow\Tests\Unit\AutoTagging;

use OCA\Workflow\AutoTagging\TaggingPlugin;
use OCA\Workflow\Tests\Unit\TestCase;
use OCP\SystemTag\TagNotFoundException;

class TaggingPluginTest extends TestCase {

	/** @var \OCP\SystemTag\ISystemTagManager|\PHPUnit\Framework\MockObject\MockObject */
	protected $manager;

	/** @var \OCP\SystemTag\ISystemTagObjectMapper|\PHPUnit\Framework\MockObject\MockObject */
	protected $mapper;

	/** @var \OCP\ILogger|\PHPUnit\Framework\MockObject\MockObject */
	protected $logger;

	/** @var \OCP\IL10N|\PHPUnit\Framework\MockObject\MockObject */
	protected $l;

	public function setUp(): void {
		parent::setUp();

		$this->manager = $this->getMockBuilder('OCP\SystemTag\ISystemTagManager')
			->disableOriginalConstructor()
			->getMock();

		$this->mapper = $this->getMockBuilder('OCP\SystemTag\ISystemTagObjectMapper')
			->disableOriginalConstructor()
			->getMock();

		$this->logger = $this->getMockBuilder('OCP\ILogger')
			->disableOriginalConstructor()
			->getMock();

		$this->l = $this->getMockBuilder('OCP\IL10N')
			->disableOriginalConstructor()
			->getMock();
	}

	/**
	 * @param array $methods
	 * @return TaggingPlugin|\PHPUnit\Framework\MockObject\MockObject
	 */
	protected function getInstance(array $methods = []) {
		if (empty($methods)) {
			return new TaggingPlugin(
				$this->manager,
				$this->mapper,
				$this->logger,
				$this->l
			);
		} else {
			return $this->getMockBuilder('OCA\Workflow\AutoTagging\TaggingPlugin')
				->setConstructorArgs([
					$this->manager,
					$this->mapper,
					$this->logger,
					$this->l,
				])
				->setMethods($methods)
				->getMock();
		}
	}

	public function dataFileAction() {
		return [
			['workflow_autotagging', ['setTags' => [23]], 42, [23], null],
			['workflow_autotagging', ['setTags' => [23]], 42, [23], new TagNotFoundException()],
			['workflow_autotagging2', null, null, null, null],
		];
	}

	/**
	 * @dataProvider dataFileAction
	 * @param string $type
	 * @param array|null $actions
	 * @param int|null $fileId
	 * @param array|null $setTags
	 * @param \Exception|null $throws
	 */
	public function testFileAction($type, $actions, $fileId, $setTags, \Exception $throws = null) {
		$plugin = $this->getInstance();

		$flow = $this->getMockBuilder('OCA\Workflow\PublicAPI\Engine\FlowInterface')
			->disableOriginalConstructor()
			->getMock();
		$flow->expects($this->once())
			->method('getType')
			->willReturn($type);
		$flow->expects($actions === null ? $this->never() : $this->once())
			->method('getActions')
			->willReturn($actions);

		$event = $this->getMockBuilder('OCA\Workflow\PublicAPI\Event\FileActionInterface')
			->disableOriginalConstructor()
			->getMock();
		$event->expects($this->once())
			->method('getFlow')
			->willReturn($flow);
		$event->expects($fileId === null ? $this->never() : $this->once())
			->method('getFileId')
			->willReturn($fileId);

		if ($throws instanceof \Exception) {
			$this->mapper->expects($this->once())
				->method('assignTags')
				->with($fileId, 'files', $setTags)
				->willThrowException($throws);
		} else {
			$this->mapper->expects($setTags === null ? $this->never() : $this->once())
				->method('assignTags')
				->with($fileId, 'files', $setTags);
		}

		/** @var \OCA\Workflow\PublicAPI\Event\FileActionInterface $event */
		$plugin->fileAction($event);
	}

	public function dataExecuteRules() {
		return [
			// No rules
			[1, [1, 24], [], []],
			// No conditions + no tags
			[1, [1, 24], [
				['conditions' => [], 'tags' => []],
			], [['tags' => [], 'result' => null]]],
			// No conditions
			[2, [2, 42], [
				['conditions' => [], 'tags' => ['test']],
			], [
				['tags' => ['test'], 'result' => null],
			]],
			// No conditions + missing tag
			[2, [2, 42], [
				['conditions' => [], 'tags' => ['test1']],
			], [
				['tags' => ['test1'], 'result' => new TagNotFoundException()],
			]],
			// No conditions + missing tag
			[2, [2, 42], [
				['conditions' => [], 'tags' => ['test1']],
				['conditions' => [], 'tags' => ['test2']],
			], [
				['tags' => ['test1'], 'result' => new TagNotFoundException()],
				['tags' => ['test2'], 'result' => null],
			]],
			[2, [2, 42], [
				['conditions' => [
					['rule' => 1, 'operator' => 'operator1', 'value' => 'value1'],
					['rule' => 2, 'operator' => 'operator2', 'value' => 'value2'],
				], 'tags' => ['test1']],
				['conditions' => [], 'tags' => ['test2']],
			], [
				['tags' => ['test2'], 'result' => null],
			]],
		];
	}
}
