<?php
/**
 * ownCloud Workflow
 *
 * @author Joas Schilling <nickvergessen@owncloud.com>
 * @copyright 2015 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Workflow\Tests\Unit\Retention;

use OCA\Workflow\Retention\Manager;
use OCA\Workflow\Retention\RetentionChecker;
use OCA\Workflow\Retention\UserBasedRetention;
use OCA\Workflow\Tests\Unit\TestCase;
use OCP\AppFramework\Utility\ITimeFactory;
use OCP\BackgroundJob\IJobList;
use OCP\Files\IRootFolder;
use OCP\Files\NotFoundException;
use OCP\IConfig;
use OCP\ILogger;
use OCP\IUser;
use OCP\IUserManager;
use OCP\IUserSession;
use OCP\SystemTag\ISystemTagObjectMapper;
use PHPUnit\Framework\MockObject\MockObject;

/**
 * Class UserBasedRetentionTest
 *
 * @group DB
 * @package OCA\Workflow\Tests\Retention
 */
class UserBasedRetentionTest extends TestCase {

	/** @var IRootFolder|MockObject */
	protected $rootFolder;

	/** @var IUserManager|MockObject */
	protected $userManager;

	/** @var ISystemTagObjectMapper|MockObject */
	protected $tagObjectMapper;

	/** @var ITimeFactory|MockObject */
	protected $timeFactory;

	/** @var IUserSession|MockObject */
	protected $userSession;

	/** @var ILogger|MockObject */
	protected $logger;

	/** @var IJobList|MockObject */
	protected $jobList;

	/** @var IConfig|MockObject */
	protected $config;

	/** @var Manager|MockObject */
	protected $retentionManager;

	/** @var RetentionChecker|MockObject */
	protected $property;

	public function setUp(): void {
		parent::setUp();

		$this->rootFolder = $this->createMock(IRootFolder::class);
		$this->userManager = $this->createMock(IUserManager::class);
		$this->tagObjectMapper = $this->createMock(ISystemTagObjectMapper::class);
		$this->timeFactory = $this->createMock(ITimeFactory::class);
		$this->userSession = $this->createMock(IUserSession::class);
		$this->logger = $this->createMock(ILogger::class);
		$this->jobList = $this->createMock(IJobList::class);
		$this->config = $this->createMock(IConfig::class);
		$this->retentionManager = $this->createMock(Manager::class);
		$this->property = $this->createMock(RetentionChecker::class);
	}

	/**
	 * @param array $methods
	 * @return UserBasedRetention|MockObject
	 */
	protected function getJob(array $methods = []) {
		if (empty($methods)) {
			return new UserBasedRetention(
				$this->rootFolder,
				$this->userManager,
				$this->tagObjectMapper,
				$this->timeFactory,
				$this->userSession,
				$this->logger,
				$this->jobList,
				$this->config,
				$this->retentionManager,
				$this->property
			);
		} else {
			return $this->getMockBuilder(UserBasedRetention::class)
				->setConstructorArgs([
					$this->rootFolder,
					$this->userManager,
					$this->tagObjectMapper,
					$this->timeFactory,
					$this->userSession,
					$this->logger,
					$this->jobList,
					$this->config,
					$this->retentionManager,
					$this->property
				])
				->setMethods($methods)
				->getMock();
		}
	}

	public function testConstruct() {
		$job = $this->getJob();
		$this->assertInstanceOf('OC\BackgroundJob\TimedJob', $job);
		$this->assertInstanceOf('OCP\BackgroundJob\IJob', $job);

		$this->assertInstanceOf('OCP\Files\IRootFolder', $this->invokePrivate($job, 'rootFolder'));
		$this->assertInstanceOf('OCP\SystemTag\ISystemTagObjectMapper', $this->invokePrivate($job, 'tagObjectMapper'));
		$this->assertInstanceOf('OCA\Workflow\Retention\Manager', $this->invokePrivate($job, 'retentionManager'));
		$this->assertInstanceOf('OCP\AppFramework\Utility\ITimeFactory', $this->invokePrivate($job, 'timeFactory'));
		$this->assertInstanceOf('OCP\IUserSession', $this->invokePrivate($job, 'userSession'));
		$this->assertInstanceOf('OCP\ILogger', $this->invokePrivate($job, 'logger'));
		$this->assertInstanceOf('OCP\BackgroundJob\IJobList', $this->invokePrivate($job, 'jobList'));
	}

	public function testConstructNoDi() {
		$job = new UserBasedRetention();
		$this->assertInstanceOf('OC\BackgroundJob\TimedJob', $job);
		$this->assertInstanceOf('OCP\BackgroundJob\IJob', $job);

		$this->assertInstanceOf('OCP\Files\IRootFolder', $this->invokePrivate($job, 'rootFolder'));
		$this->assertInstanceOf('OCP\SystemTag\ISystemTagObjectMapper', $this->invokePrivate($job, 'tagObjectMapper'));
		$this->assertInstanceOf('OCA\Workflow\Retention\Manager', $this->invokePrivate($job, 'retentionManager'));
		$this->assertInstanceOf('OCP\AppFramework\Utility\ITimeFactory', $this->invokePrivate($job, 'timeFactory'));
		$this->assertInstanceOf('OCP\IUserSession', $this->invokePrivate($job, 'userSession'));
		$this->assertInstanceOf('OCP\ILogger', $this->invokePrivate($job, 'logger'));
		$this->assertInstanceOf('OCP\BackgroundJob\IJobList', $this->invokePrivate($job, 'jobList'));
	}

	public function dataRun() {
		return [
			['user1', false, null, null, false],
			['admin', true, '0', null, false],
			['admin', true, '1', '10', false],
			['admin', true, '1', '-10', true],
		];
	}

	/**
	 * @dataProvider dataRun
	 *
	 * @param string $uid
	 * @param bool $userExists
	 * @param int $folderRetention
	 * @param int $folderRetentionPeriod
	 * @param bool $folderException
	 */
	public function testRun($uid, $userExists, $folderRetention, $folderRetentionPeriod, $folderException) {
		$job = $this->getJob([
			'getRetentionPeriods',
			'getRetentionTimestamp',
			'getUserFolder',
			'executeRetention',
		]);

		$tags = $foldersInQueue = [];
		if ($userExists) {
			$user = $this->createMock(IUser::class);
			$this->userManager->expects($this->once())
				->method('get')
				->with($uid)
				->willReturn($user);

			$job->expects($this->once())
				->method('getRetentionPeriods');

			$job->expects($this->once())
				->method('executeRetention')
				->with($user);

			if (!$folderException) {
				$job->expects($folderRetentionPeriod === null ? $this->never() : $this->once())
					->method('getRetentionTimestamp')
					->with([
						'unit' => 'days',
						'numUnits' => (int) $folderRetentionPeriod,
					]);
			} else {
				$job->expects($this->once())
					->method('getRetentionTimestamp')
					->with([
						'unit' => 'days',
						'numUnits' => (int) $folderRetentionPeriod,
					])
				->willThrowException(new \OutOfBoundsException());
			}

			$userFolder = $this->getNodeMock(12, 'Folder');
			$job->expects($this->once())
				->method('getUserFolder')
				->with($user)
				->willReturn($userFolder);
			$foldersInQueue = [$userFolder];
			$tags = [12 => []];

			$this->config->expects($this->atLeastOnce())
				->method('getAppValue')
				->willReturnMap([
					['workflow', 'folder_retention', false, $folderRetention],
					['workflow', 'folder_retention_period', 7, $folderRetentionPeriod],
				]);

			$this->jobList->expects($this->never())
				->method('remove');

			$this->logger->expects($this->never())
				->method('warning');

			$this->logger->expects($folderException ? $this->once() : $this->never())
				->method('error')
				->with('Invalid folder retention period "{config}".', [
					'config' => $folderRetentionPeriod,
					'app' => 'workflow',
				]);
		} else {
			$this->userManager->expects($this->once())
				->method('get')
				->with($uid)
				->willReturn(null);

			$job->expects($this->never())
				->method('getRetentionPeriods');

			$this->config->expects($this->never())
				->method('getAppValue');

			$job->expects($this->never())
				->method('executeRetention');

			$job->expects($this->never())
				->method('getUserFolder');

			$this->jobList->expects($this->once())
				->method('remove')
				->with('\OCA\Workflow\Retention\UserBasedRetention', ['user' => $uid]);

			$this->logger->expects($this->once())
				->method('warning')
				->with('Removing retention for user "{user}" because the user could not be found.', [
					'user' => $uid,
					'app' => 'workflow',
				]);
		}

		$this->invokePrivate($job, 'run', [['user' => $uid]]);

		$this->assertEquals($foldersInQueue, $this->invokePrivate($job, 'queueListing'));
		$this->assertEquals($tags, $this->invokePrivate($job, 'tags'));
	}

	protected function getNodeMock($id = 0, $class = 'Node', $isShared = false) {
		$node = $this->getMockBuilder('OCP\\Files\\' . $class)
			->disableOriginalConstructor()
			->getMock();

		if (\is_numeric($id) && $id) {
			$node->expects($this->any())
				->method('getId')
				->willReturn($id);
		} elseif (!\is_numeric($id)) {
			$node->expects($this->any())
				->method('getId')
				->willThrowException(new \Exception());
		}

		$storage = $this->getMockBuilder('OCP\Files\Storage')
			->disableOriginalConstructor()
			->getMock();
		$storage->expects($this->any())
			->method('instanceOfStorage')
			->with('OC\Files\Storage\Shared')
			->willReturn($isShared);

		$node->expects($this->any())
			->method('getStorage')
			->willReturn($storage);

		return $node;
	}

	public function dataExecuteRetention() {
		return [
			[false, false, true, null, null, null, true],
			[false, true, true, null, null, null, true],
			[true, false, true, null, null, null, false],

			[false, false, false, 0, 123456789, null, true],
			[false, true, false, 0, 123456789, null, true],
			[true, false, false, 0, 123456789, null, false],

			[false, false, false, 12345678, 123456789, null, true],
			[false, false, false, 123456789, 12345678, false, true],
			[false, false, false, 123456789, 12345678, true, false],
			[true, false, false, 123456789, null, null, false],
		];
	}

	/**
	 * @dataProvider dataExecuteRetention
	 * @param bool $getDirectoryThrows
	 * @param bool $hasChildren
	 * @param bool $folderIsRoot
	 * @param int|null $folderRetention
	 * @param int|null $folderMTime
	 * @param bool|null $deleteFolderThrows
	 * @param bool $expected
	 */
	public function testExecuteRetention($getDirectoryThrows, $hasChildren, $folderIsRoot, $folderRetention, $folderMTime, $deleteFolderThrows, $expected) {
		$job = $this->getJob([
			'addNodeToQueue',
			'processQueue',
		]);

		$user = $this->createMock('OCP\IUser');
		$user->expects($this->any())
			->method('getUID')
			->willReturn('user');

		$folder = $this->getNodeMock(42, 'Folder');

		if ($getDirectoryThrows) {
			$exception = new NotFoundException();
			$folder->expects($this->once())
				->method('getDirectoryListing')
				->willThrowException($exception);

			$job->expects($this->never())
				->method('addNodeToQueue');

			$job->expects($this->never())
				->method('processQueue');

			$this->logger->expects($this->once())
				->method('logException')
				->with($exception, $this->anything());
		} elseif ($hasChildren) {
			$node1 = $this->getNodeMock();
			$node2 = $this->getNodeMock();

			$folder->expects($this->once())
				->method('getDirectoryListing')
				->willReturn([$node1, $node2]);

			$job->expects($this->exactly(2))
				->method('addNodeToQueue')
				->withConsecutive(
					[$node1, $folder],
					[$node2, $folder]
				);

			$job->expects($this->once())
				->method('processQueue');
		} else {
			$folder->expects($this->once())
				->method('getDirectoryListing')
				->willReturn([]);

			$job->expects($this->never())
				->method('addNodeToQueue');

			$job->expects($this->never())
				->method('processQueue');
		}

		if ($folderIsRoot) {
			$userRoot = $folder;
		} else {
			$userRoot = $this->getNodeMock(23, 'Folder');

			$this->property->expects(($hasChildren || !$folderRetention || $folderMTime === null) ? $this->never() : $this->once())
				->method('isRetentionOver')
				->willReturn($folderMTime < $folderRetention);

			if ($deleteFolderThrows === null) {
				$folder->expects($this->never())
					->method('delete');
			} elseif ($deleteFolderThrows) {
				$deleteException = new \Exception();
				$folder->expects($this->once())
					->method('delete')
					->willThrowException($deleteException);

				$this->logger->expects($this->once())
					->method('logException')
					->with($deleteException, $this->anything());
			} else {
				$folder->expects($this->once())
					->method('delete');
			}
		}

		// We must NEVER compare the mtime of the user's root folder,
		// because this means we consider deleting it...
		$userRoot->expects($this->never())
			->method('getMTime');

		$this->invokePrivate($job, 'folderRetentionPeriod', [$folderRetention]);
		$this->invokePrivate($job, 'queueListing', [[$folder]]);
		$this->invokePrivate($job, 'userFolder', [$userRoot]);

		$this->assertSame($expected, $this->invokePrivate($job, 'executeRetention', [$user]));
	}

	public function dataAddNodeToQueue() {
		return [
			[0, UserBasedRetention::BATCH_SIZE + 1, false, false, false],
			[1, UserBasedRetention::BATCH_SIZE + 1, false, false, false],
			[UserBasedRetention::BATCH_SIZE - 2, UserBasedRetention::BATCH_SIZE + 1, false, false, false],
			[UserBasedRetention::BATCH_SIZE - 1, UserBasedRetention::BATCH_SIZE + 1, false, true, true],
			[UserBasedRetention::BATCH_SIZE, UserBasedRetention::BATCH_SIZE + 1, false, true, true],
			[0, true, false, false, false],
			[UserBasedRetention::BATCH_SIZE, true, false, false, false],

			[UserBasedRetention::BATCH_SIZE, UserBasedRetention::BATCH_SIZE + 1, true, false, false],
		];
	}

	/**
	 * @dataProvider dataAddNodeToQueue
	 * @param int $exitingJobs
	 * @param int|true $newNodeId Integer or true for $throws
	 * @param bool $isShared
	 * @param bool $runsProcess
	 * @param bool $expected
	 */
	public function testAddNodeToQueue($exitingJobs, $newNodeId, $isShared, $runsProcess, $expected) {
		$job = $this->getJob(['processQueue']);

		$nodes = [];
		for ($i = 1; $i <= $exitingJobs; $i++) {
			$nodes[$i] = $this->getNodeMock($i);
		}
		$this->invokePrivate($job, 'queueRetention', [$nodes]);

		$job->expects($runsProcess ? $this->once() : $this->never())
			->method('processQueue');

		$addNode = $this->getNodeMock($newNodeId, 'Node', $isShared);
		$parent = $this->getNodeMock(1337, 'Folder');

		$this->assertSame($expected, $this->invokePrivate($job, 'addNodeToQueue', [$addNode, $parent]));
	}

	public function testProcessQueueEmpty() {
		$job = $this->getJob();

		$this->invokePrivate($job, 'queueRetention', [[]]);

		$this->tagObjectMapper->expects($this->never())
			->method('getTagIdsForObjects');

		$this->invokePrivate($job, 'processQueue');
	}

	public function dataProcessQueue() {
		return [
			[[1 => ['type' => 'File']], [], 0],
			[[1 => ['type' => 'Folder', 'listingQueue' => true]], [], 0],

			[[1 => ['type' => 'File']], [1 => []], 0],
			[[1 => ['type' => 'File', 'getMTime' => 123456789]], [1 => [123456788]], 0],
			[
				[
					1 => ['type' => 'Folder', 'getMTime' => 123456789, 'delete' => 'throws'],
					2 => ['type' => 'Folder', 'getMTime' => 123456789, 'delete' => true],
				], [
					1 => [123456790],
					2 => [123456791],
				],
				1,
			],
		];
	}

	/**
	 * @dataProvider dataProcessQueue
	 * @param array $nodes
	 * @param array $tags
	 * @param int $numExceptionsLogged
	 */
	public function testProcessQueue($nodes, $tags, $numExceptionsLogged) {
		$job = $this->getJob(['getRetentionPeriodForTag']);
		$job->expects($this->any())
			->method('getRetentionPeriodForTag')
			->willReturnArgument(0);

		$queue = [];
		$queueListing = [];
		foreach ($nodes as $id => $data) {
			$node = $this->getNodeMock($id, $data['type']);
			if (!empty($data['listingQueue'])) {
				$queueListing[] = $node;
			}
			if (!empty($data['getMTime'])) {
				$this->property
					->method('isRetentionOver')
					->willReturnCallback(
						function ($node, $timestamp) use ($data) {
							return $data['getMTime'] < $timestamp;
						}
					);
			} else {
				$this->property->expects($this->never())
					->method('isRetentionOver');
			}
			if (!empty($data['delete'])) {
				if (\is_bool($data['delete'])) {
					$node->expects($this->once())
						->method('delete');
				} else {
					$node->expects($this->once())
						->method('delete')
						->willThrowException(new \Exception());
				}
			} else {
				$node->expects($this->never())
					->method('delete');
			}
			$queue[$id] = ['node' => $node, 'parent' => isset($data['parent']) ? $data['parent'] : 0];
		}
		$this->invokePrivate($job, 'queueRetention', [$queue]);
		$this->invokePrivate($job, 'tags', [[0 => []]]);

		$this->tagObjectMapper->expects($this->once())
			->method('getTagIdsForObjects')
			->with(\array_keys($queue), 'files')
			->willReturn($tags);

		$this->logger->expects($this->exactly($numExceptionsLogged))
			->method('logException');

		$this->invokePrivate($job, 'processQueue');

		$this->assertEquals($queueListing, $this->invokePrivate($job, 'queueListing'));
	}

	public function dataGetRetentionPeriodForTag() {
		return [
			['tag', null, [], false],
			['tag', null, ['tag' => 123], 123],
			['tag', [], null, false],
			['tag', ['tag' => 123], null, 123],
		];
	}

	/**
	 * @dataProvider dataGetRetentionPeriodForTag
	 * @param string $tag
	 * @param null|array $prefilledRetentions
	 * @param array $loadedRetentions
	 * @param false|int $expected
	 */
	public function testGetRetentionPeriodForTag($tag, $prefilledRetentions, $loadedRetentions, $expected) {
		$job = $this->getJob(['getRetentionPeriods']);

		if ($prefilledRetentions !== null) {
			$this->invokePrivate($job, 'retentionPeriods', [$prefilledRetentions]);
		}
		if ($loadedRetentions !== null) {
			$job->expects($this->once())
				->method('getRetentionPeriods')
				->willReturn($loadedRetentions);
		}

		$this->assertSame($expected, $this->invokePrivate($job, 'getRetentionPeriodForTag', [$tag]));
	}

	public function testGetRetentionPeriods() {
		$job = $this->getJob();

		$this->timeFactory->expects($this->atLeastOnce())
			->method('getTime')
			->willReturn(1414195200); // 2014-10-25T00:00:00+00:00

		$this->retentionManager->expects($this->once())
			->method('getAll')
			->willReturn([
				[
					'tagId' => 'one',
					'numUnits' => (int) 1,
					'unit' => 'days',
				],
				[
					'tagId' => 'two',
					'numUnits' => 'two',
					'unit' => 'days',
				],
				[
					'tagId' => 'three',
					'numUnits' => -3,
					'unit' => 'days',
				],
				[
					'tagId' => 'four',
					'numUnits' => 4,
					'unit' => 'no-unit',
				],
				[
					'tagId' => 'five',
					'numUnits' => 5,
					'unit' => 'months',
				],
			]);

		$this->assertSame([
			'one' => 1414108800, // 2014-10-24T00:00:00+00:00
			'five' => 1400976000, // 2014-05-25T00:00:00+00:00
		], $this->invokePrivate($job, 'getRetentionPeriods'));
	}

	public function testGetRetentionPeriodsLoaded() {
		$job = $this->getJob();

		$this->invokePrivate($job, 'retentionPeriods', [[
			'one' => 1414108800,
			'five' => 1400976000,
		]]);

		$this->assertSame([
			'one' => 1414108800,
			'five' => 1400976000,
		], $this->invokePrivate($job, 'getRetentionPeriods'));
	}

	public function dataGetRetentionTimestamp() {
		return [
			[7, 'days', 1414195200], // 2014-10-25T00:00:00+00:00
			[1, 'weeks', 1414195200], // 2014-10-25T00:00:00+00:00
			[3, 'months', 1406851200], // 2014-08-01T00:00:00+00:00
			[20, 'years', 783648000], // 1994-11-01T00:00:00+00:00
		];
	}

	/**
	 * @dataProvider dataGetRetentionTimestamp
	 * @param int $numUnits
	 * @param string $unit
	 * @param int $expected
	 */
	public function testGetRetentionTimestamp($numUnits, $unit, $expected) {
		$job = $this->getJob();

		$this->timeFactory->expects($this->once())
			->method('getTime')
			->willReturn(1414800000);// 2014-11-01T00:00:00+00:00

		$this->assertSame($expected, $this->invokePrivate($job, 'getRetentionTimestamp', [[
			'numUnits' => $numUnits,
			'unit' => $unit,
		]]));
	}

	public function dataGetRetentionTimestampInvalid() {
		return [
			['7', 'days', 'Invalid retention period for tag "{tag}"', 1],
			['-7', 'days', 'Invalid retention period for tag "{tag}"', 1],
			[-7, 'days', 'Invalid retention period for tag "{tag}"', 1],
			[1, 'hours', 'Invalid retention period unit for tag "{tag}"', 2],
			[3, 'minutes', 'Invalid retention period unit for tag "{tag}"', 2],
			[20, 'seconds', 'Invalid retention period unit for tag "{tag}"', 2],
		];
	}

	/**
	 * @dataProvider dataGetRetentionTimestampInvalid
	 * @param int $numUnits
	 * @param string $unit
	 * @param string $message
	 * @param int $code
	 */
	public function testGetRetentionTimestampInvalid($numUnits, $unit, $message, $code) {
		$job = $this->getJob();

		$this->timeFactory->expects($this->never())
			->method('getTime');

		$this->expectException('OutOfBoundsException');
		$this->expectExceptionMessage($message);
		$this->expectExceptionCode($code);

		$this->invokePrivate($job, 'getRetentionTimestamp', [[
			'numUnits' => $numUnits,
			'unit' => $unit,
		]]);
	}
}
